/*	
 *	FIXME: does not work on big-endian machines
 */

#define _POSIX_C_SOURCE 200112L

#include <stdlib.h>
#include <iconv.h>
#include <stdio.h>
#include <errno.h>
#include <stdint.h>
#include <stdbool.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>

typedef struct {
	uint8_t *c;
	size_t size;
} str_t;

static const uint16_t tagTableEmptyField = 0xffff;

typedef struct {
	uint16_t flags1, flags2, flags3, flags4;
	uint8_t unknown2[11];
	str_t path;
	str_t filename;
	str_t title;
	str_t artist;
	str_t album;
	str_t genre;
	str_t year;
	uint8_t unknown3[2];
	uint8_t cd;
	uint8_t track;
} tagTableEntry_t;

typedef struct {
	const char *data;
	size_t offset;
	size_t headerSize;
	size_t size;
	size_t count;
	size_t musicCount;
	size_t unknownCount1;
	size_t audiobooksCount;
	size_t recordingsCount;
	size_t unknownCount2;
	size_t podcastsCount;
} tagTable_t;

typedef uint16_t lookupTableIdx_t;
static const lookupTableIdx_t lookupTableInvalidIdx = 0xfffe;

typedef struct {
	uint32_t tagOffset;
	uint8_t flags;
	lookupTableIdx_t unknown;
	lookupTableIdx_t next;
} lookupTableEntry_t;

typedef struct {
	const char *data;
	size_t offset;
	size_t headerSize;
	size_t size;
	size_t count;
	lookupTableIdx_t firstArtist;
	lookupTableIdx_t firstAlbum;
	lookupTableIdx_t firstTitle;
	lookupTableIdx_t firstGenre;
	lookupTableIdx_t unknown1[2];
	lookupTableIdx_t firstAudiobook;
	lookupTableIdx_t firstRecording;
	lookupTableIdx_t unknown2[2];
	lookupTableIdx_t mostRecentTitle;
	lookupTableIdx_t firstPodcast;
	lookupTableIdx_t firstFolder;
	lookupTableIdx_t unknown3[2];
} lookupTable_t;

typedef struct {
	const char *data;
	size_t size;
	tagTable_t tagTable;
	lookupTable_t lookupTable;
} mtable_t;

static void strFree (str_t *str) {
	assert (str != NULL);

	free (str->c);
	memset (str, 0, sizeof (*str));
}

static const void *tagTableGetFieldEnd (const char *data, size_t size) {
	assert ((uintptr_t) data % 2 == 0);

	while (size > sizeof (uint16_t)-1) {
		if (*((uint16_t *) data) == 0x0000) {
			return data;
		}
		data += sizeof (uint16_t);
		size -= sizeof (uint16_t);
	}
	return NULL;
}

static void tagTableCopyStrField (const char *data, size_t size, str_t *str) {
	const char *end;
	end = tagTableGetFieldEnd (data, size);
	assert (end != NULL);
	/* FIXME: size */
	if (end - data == 2 && *((uint16_t *) data) == tagTableEmptyField) {
		str->size = 0;
		str->c = NULL;
//		fprintf (stderr, "found empty field\n");
		return;
	}
//	fprintf (stderr, "found field with length %zu, content: %s\n", end - data, data);
	str->size = end - data;
	assert (str->size % 2 == 0);
	str->c = malloc (str->size * sizeof (*str->c));
	memcpy (str->c, data, str->size);
}

static void tagTableFreeEntry (tagTableEntry_t *entry) {
	strFree (&entry->path);
	strFree (&entry->filename);
	strFree (&entry->title);
	strFree (&entry->artist);
	strFree (&entry->album);
	strFree (&entry->genre);
	strFree (&entry->year);

	memset (entry, 0, sizeof (*entry));
}

static size_t tagTableReadEntry (const char *data, const size_t size,
		tagTableEntry_t *entry) {
	const char *dataPos = data;

	/* to make things easier we assume data address is even */
	assert ((uintptr_t) data % 2 == 0);

	assert (*((uint16_t *) dataPos) == 0x0000);
	dataPos += 2;

	entry->flags1 = *((uint16_t *) dataPos);
	dataPos += 2;
	entry->flags2 = *((uint16_t *) dataPos);
	dataPos += 2;
	entry->flags3 = *((uint16_t *) dataPos);
	dataPos += 2;
	entry->flags4 = *((uint16_t *) dataPos);
	dataPos += 2;

	assert (memcmp (dataPos, "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00", 10) == 0);
	dataPos += 10;

	/* FIXME: size! */
	str_t cdAndTrack;
	str_t *strFields[] = {&entry->path, &entry->filename, &entry->title,
			&entry->artist, &entry->album, &entry->genre, &entry->year, &cdAndTrack};
	for (size_t i = 0; i < sizeof (strFields) / sizeof (*strFields); i++) {
		str_t *field = strFields[i];
		tagTableCopyStrField (dataPos, size, field);
		if (field->size == 0) {
			dataPos += sizeof (tagTableEmptyField);
		}
		dataPos += field->size + 2;
	}

	/* FIXME: move cd and track # to datastruct */
	strFree (&cdAndTrack);

	return dataPos - data;
}

/*	length of tag table is not part of offset (and not part of the table at
 *	all?)
 */
inline static const char *tagTableTranslateOffset (const tagTable_t *tt,
		size_t offset) {
	return tt->data + offset + 4;
}

static void tagTableInit (tagTable_t *tt, const char *data, const char size) {
	tt->offset = 0x0L;
	tt->headerSize = 32;
	tt->data = data + tt->offset;

	const char *dataPos = tt->data;

	tt->size = *((uint32_t *) dataPos);
	dataPos += sizeof (uint32_t);

	size_t *counts[] = {&tt->musicCount, &tt->unknownCount1,
			&tt->audiobooksCount, &tt->recordingsCount, &tt->unknownCount2,
			&tt->podcastsCount};
	for (size_t i = 0; i < sizeof (counts) / sizeof (*counts); i++) {
		*counts[i] = *((uint32_t *) dataPos);
		dataPos += sizeof (uint32_t);
	}
	assert (tt->unknownCount1 == 0x0L);
	assert (tt->unknownCount2 == 0x0L);
}

static size_t lookupTableReadEntry (const char *data, const size_t size,
		lookupTableEntry_t *entry) {
	const char *dataPos = data;

	entry->tagOffset = *((uint32_t *) dataPos) & 0x00ffffff;
	dataPos += 3;

	entry->flags = *((uint8_t *) dataPos);
	dataPos += 1;

	entry->unknown = *((lookupTableIdx_t *) dataPos);
	dataPos += 2;

	entry->next = *((lookupTableIdx_t *) dataPos);
	dataPos += 2;

	return dataPos - data;
}

inline static const char *lookupTableIndexToAddress (const lookupTable_t *lt,
		const lookupTableIdx_t idx) {
	if (idx == lookupTableInvalidIdx) {
		return NULL;
	} else {
		return lt->data + lt->headerSize + 8 * idx;
	}
}

static void lookupTableInit (lookupTable_t *lt, const char *data,
		const char size) {
	lt->offset = 0x0013a104L;
	lt->headerSize = 64;
	lt->data = data + lt->offset;

	const char *dataPos = lt->data;
	lookupTableIdx_t *ids[] = {&lt->firstArtist, &lt->firstAlbum,
			&lt->firstTitle, &lt->firstGenre, &lt->unknown1[0],
			&lt->unknown1[1], &lt->unknown1[2], &lt->firstRecording,
			&lt->unknown2[0], &lt->unknown2[1], &lt->mostRecentTitle,
			&lt->firstPodcast, &lt->firstFolder, &lt->unknown3[0],
			&lt->unknown3[1]};
	for (size_t i = 0; i < sizeof (ids) / sizeof (*ids); i++) {
		*ids[i] = *((uint16_t *) dataPos);
		dataPos += 2;
		/* skip nul */
		dataPos += 2;
	}

	lt->count = *((uint16_t *) dataPos);
	dataPos += 2;

	assert (lt->unknown1[0] == lookupTableInvalidIdx);
	assert (lt->unknown1[1] == lookupTableInvalidIdx);
	assert (lt->unknown2[0] == lookupTableInvalidIdx);
	assert (lt->unknown2[1] == lookupTableInvalidIdx);
	assert (lt->unknown3[0] == lookupTableInvalidIdx);
	assert (lt->unknown3[1] == lookupTableInvalidIdx);

	lt->size = lt->count * 8 + lt->headerSize;
}

static void mtableInit (mtable_t *mt, const char *data, const size_t size) {
	mt->data = data;
	mt->size = size;
	tagTableInit (&mt->tagTable, data, size);
	lookupTableInit (&mt->lookupTable, data, size);
}

static bool mmapFile (const char *file, char **retContent, size_t *retSize) {
	char *content;
	size_t contentSize;
	int fd;
	struct stat sb;

	fd = open (file, O_RDONLY);
	assert (fd != -1);

	fstat (fd, &sb);
	contentSize = sb.st_size;

	content = mmap (NULL, contentSize, PROT_READ, MAP_PRIVATE, fd, 0);
	assert (content != MAP_FAILED);

	close (fd);

	*retContent = content;
	*retSize = contentSize;

	return true;
}

static void iconvWrite (int fd, iconv_t ic, char *data, size_t size) {
	char out[1024], *outpos = out;
	size_t outsize = sizeof (out);
	size_t ret;
	int iconvErrno;

	do {
		ret = iconv (ic, &data, &size, &outpos, &outsize);
		iconvErrno = errno;

		write (fd, out, sizeof (out) - outsize);
		outpos = out;
		outsize = sizeof (out);
	} while (ret == -1 && iconvErrno == E2BIG);
}

static void dumpLookupStructure (iconv_t ic, const mtable_t *mt, lookupTableIdx_t idx,
		size_t indent) {
	const char *dataPos;

	while (idx != lookupTableInvalidIdx) {
		dataPos = lookupTableIndexToAddress (&mt->lookupTable, idx);

		lookupTableEntry_t entry;
		lookupTableReadEntry (dataPos, mt->lookupTable.size, &entry);

		str_t tagValue, tagValue2;
		dataPos = tagTableTranslateOffset (&mt->tagTable, entry.tagOffset);
		tagTableCopyStrField (dataPos, mt->tagTable.size, &tagValue);
		if (tagValue.size == 0) {
			dataPos += sizeof (tagTableEmptyField);
		}
		dataPos += tagValue.size + 2;
		tagTableCopyStrField (dataPos, mt->tagTable.size, &tagValue2);

		for (size_t i = 0; i < indent; i++) {
			fputc (' ', stdout);
		}
		fprintf (stdout, "%04x: flags: %02x, unknown: %04x, next: %04x, tag: ", idx, entry.flags, entry.unknown, entry.next);
		fflush (stdout);
		if (entry.flags & 0x40 || entry.flags == 0x01 || entry.flags & 0x80) {
			write (STDOUT_FILENO, (char *) tagValue.c, tagValue.size);
			if (entry.flags & 0x80) {
				write (STDOUT_FILENO, (char *) tagValue2.c, tagValue2.size);
			}
		} else {
			iconvWrite (STDOUT_FILENO, ic, (char *) tagValue.c, tagValue.size);
		}
		write (STDOUT_FILENO, "\n", 1);

		/* seems to indicate "end of list"/"do not descend" */
		if (!(entry.flags & 0x80)) {
			dumpLookupStructure (ic, mt, entry.unknown, indent+1);
		}

		strFree (&tagValue);

		if (entry.flags & 0x80) {
			idx = lookupTableInvalidIdx;
		} else {
			idx = entry.next;
		}
	}
}

int main (int argc, char **argv) {
	char *data;
	size_t size;
	mtable_t mt;
	iconv_t ic = iconv_open ("UTF-8", "UTF-16");

	mmapFile (argv[1], &data, &size);

	mtableInit (&mt, data, size);
	fprintf (stderr, "initialized mtable at %p, size %zu (%zx)\n", data, size,
			size);
	fprintf (stderr, "tag table is at %p, size %zu (%zx)\n",
			mt.tagTable.data, mt.tagTable.size, mt.tagTable.size);
	fprintf (stderr, "lookup table is at %p, size %zu (%zx), entries %zu (%zx)\n",
			mt.lookupTable.data, mt.lookupTable.size, mt.lookupTable.size,
			mt.lookupTable.count, mt.lookupTable.count);

	/* dump tag table */
	size_t dumpedTags = 0;
	const char *dataPos = mt.tagTable.data + mt.tagTable.headerSize;
	while (dataPos < mt.tagTable.data + mt.tagTable.size) {
		tagTableEntry_t entry;
		size_t consumed = tagTableReadEntry (dataPos,
				mt.tagTable.size - mt.tagTable.headerSize, &entry);

		fprintf (stdout, "%04x %04x %04x %04x ", entry.flags1, entry.flags2, entry.flags3, entry.flags4);
		fflush (stdout);
		write (STDOUT_FILENO, entry.path.c, entry.path.size);
		write (STDOUT_FILENO, entry.filename.c, entry.filename.size);
		write (STDOUT_FILENO, "\n", 1);

#if 0
		iconvWrite (STDOUT_FILENO, ic, (char *) entry.artist.c, entry.artist.size);
		write (STDOUT_FILENO, " - ", 3);
		iconvWrite (STDOUT_FILENO, ic, (char *) entry.title.c, entry.title.size);
		write (STDOUT_FILENO, "\n", 1);
#endif

		dataPos += consumed;
		++dumpedTags;

		tagTableFreeEntry (&entry);
	}

	fprintf (stderr, "dumped %zu tags\n", dumpedTags);

	fprintf (stderr, "most recent title is at idx %u (%x)\n", mt.lookupTable.mostRecentTitle, mt.lookupTable.mostRecentTitle);
	dumpLookupStructure (ic, &mt, mt.lookupTable.firstGenre, 0);

	iconv_close (ic);

	munmap (data, size);
}

